@@ -25,6 +25,7 @@
 
 #include "fuse-overlayfs.h"
 
+#include "limits.h"
 #include <unistd.h>
 #include <fcntl.h>
 #include <errno.h>
@@ -78,8 +79,8 @@ direct_getxattr (struct ovl_layer *l, const char *path, const char *name, char *
 static int
 direct_fstat (struct ovl_layer *l, int fd, const char *path, unsigned int mask, struct stat *st)
 {
-#ifdef HAVE_STATX
   int ret;
+#ifdef HAVE_STATX
   struct statx stx;
 
   ret = statx (fd, "", AT_STATX_DONT_SYNC|AT_EMPTY_PATH, mask, &stx);
@@ -87,21 +88,27 @@ direct_fstat (struct ovl_layer *l, int fd, const char *path, unsigned int mask,
   if (ret < 0 && errno == ENOSYS)
     goto fallback;
   if (ret == 0)
-    statx_to_stat (&stx, st);
+    {
+      statx_to_stat (&stx, st);
+      return override_mode (l, fd, NULL, path, st);
+    }
 
   return ret;
 #endif
 
  fallback:
-  return fstat (fd, st);
+  ret = fstat (fd, st);
+  if (ret != 0)
+    return ret;
 
+  return override_mode (l, fd, NULL, path, st);
 }
 
 static int
 direct_statat (struct ovl_layer *l, const char *path, struct stat *st, int flags, unsigned int mask)
 {
-#ifdef HAVE_STATX
   int ret;
+#ifdef HAVE_STATX
   struct statx stx;
 
   ret = statx (l->fd, path, AT_STATX_DONT_SYNC|flags, mask, &stx);
@@ -109,12 +116,19 @@ direct_statat (struct ovl_layer *l, const char *path, struct stat *st, int flags
   if (ret < 0 && errno == ENOSYS)
     goto fallback;
   if (ret == 0)
-    statx_to_stat (&stx, st);
+    {
+      statx_to_stat (&stx, st);
+      return override_mode (l, -1, NULL, path, st);
+    }
 
   return ret;
 #endif
  fallback:
-  return fstatat (l->fd, path, st, flags);
+  ret = fstatat (l->fd, path, st, flags);
+  if (ret != 0)
+    return ret;
+
+  return override_mode (l, -1, NULL, path, st);
 }
 
 static struct dirent *
@@ -129,7 +143,7 @@ direct_opendir (struct ovl_layer *l, const char *path)
   cleanup_close int cleanup_fd = -1;
   DIR *dp = NULL;
 
-  cleanup_fd = TEMP_FAILURE_RETRY (openat (l->fd, path, O_DIRECTORY));
+  cleanup_fd = TEMP_FAILURE_RETRY (safe_openat (l->fd, path, O_DIRECTORY, 0));
   if (cleanup_fd < 0)
     return NULL;
 
@@ -151,7 +165,7 @@ direct_closedir (void *dirp)
 static int
 direct_openat (struct ovl_layer *l, const char *path, int flags, mode_t mode)
 {
-  return TEMP_FAILURE_RETRY (openat (l->fd, path, flags, mode));
+  return TEMP_FAILURE_RETRY (safe_openat (l->fd, path, flags, mode));
 }
 
 static ssize_t
@@ -163,6 +177,7 @@ direct_readlinkat (struct ovl_layer *l, const char *path, char *buf, size_t bufs
 static int
 direct_load_data_source (struct ovl_layer *l, const char *opaque, const char *path, int n_layer)
 {
+  char tmp[64];
   l->path = realpath (path, NULL);
   if (l->path == NULL)
     {
@@ -178,6 +193,13 @@ direct_load_data_source (struct ovl_layer *l, const char *opaque, const char *pa
       return l->fd;
     }
 
+  if (fgetxattr (l->fd, XATTR_PRIVILEGED_OVERRIDE_STAT, tmp, sizeof (tmp)) >= 0)
+    l->stat_override_mode = STAT_OVERRIDE_PRIVILEGED;
+  else if (fgetxattr (l->fd, XATTR_OVERRIDE_STAT, tmp, sizeof (tmp)) >= 0)
+    l->stat_override_mode = STAT_OVERRIDE_USER;
+  else if (fgetxattr (l->fd, XATTR_OVERRIDE_CONTAINERS_STAT, tmp, sizeof (tmp)) >= 0)
+    l->stat_override_mode = STAT_OVERRIDE_CONTAINERS;
+
   return 0;
 }
 
